'''Containing module for the Calculator class.'''

import logging
from typing import List, Dict, Union
from math import pi, e
from veq.tokens import TokenBuilder, Token, TokenStream, VariableToken, FunctionToken

class CalculationError(Exception):

    '''An exception that indicates there was an error in calculating the user's equation.'''
    
    def __init__(self, expression: str):
        self.__expression = expression
        super().__init__(expression)

    def __str__(self):
        return f"Error in equation \"f(x) = {self.__expression}\""

class VariableUndefinedError(Exception):

    '''An exception that indicates the user used an undefined variable.'''

    def __init__(self, variable: str):
        self.__variable = variable
        super().__init__(variable)

    def __str__(self):
        return f"Variable undefined: \"{self.__variable}\""

class Calculator:

    '''Class responsible for converting infix to postfix and calculating expressions.

    :attribute stream: Expression in infix format for reading expressions.
    :type stream: str or TokenStream
    :attribute stack: Calculation stack.
    :type stack: List(float)
    :value stack: []'''

    CONSTANTS = {
        "pi": pi,
        "e": e,
        "g": 9.80665
    }

    def __init__(self, stream: Union[str, TokenStream], stack: List[float] = []):
        self.stack = stack
        self.__builder = TokenBuilder(self.stack)

        if isinstance(stream, str):
            self.stream = TokenStream(stream)
        else:
            self.stream = stream

        self.expression = []

    def infix_to_postfix(self):
        '''Convert infix stream to postfix list of tokens.

        :param expression: String to tokenize.
        :type expression: str
        :returns: List of tokens
        :rtype: list(Token)'''

        lookup = TokenBuilder.TOKENS

        stack: List[Token] = []

        token: Token = None

        for match in self.stream:
            if match in lookup:
                # This is really wonky. Basically: Look up what the name of function is.
                function_name = lookup[match]
                # Get the function with that name from the instance of our class.
                build_function = self.__builder.__getattribute__(function_name)
                # Execute that function.
                token = build_function()
            elif match == '(':
                self.infix_to_postfix()
                continue
            elif match == ')':
                while len(stack) > 0:
                    self.expression.append(stack.pop())
                return
            elif match.isalpha():
                token = self.__builder.build_variable(name = match)
                self.expression.append(token)
                continue
            else:
                token = self.__builder.build_value(value = float(match))
                self.expression.append(token)
                continue

            if isinstance(token, FunctionToken):
                self.infix_to_postfix()
                self.expression.append(token)
                continue

            while len(stack) > 0 and stack[-1].precedence >= token.precedence:
                self.expression.append(stack.pop())
            stack.append(token)

        while len(stack) > 0:
            self.expression.append(stack.pop())

    def calculate(self, **variables) -> float:
        '''Calculate the Calculator's expression.

        :param variables: Values to substitute for x in the expression.
        :type variables: dict
        :returns: The result of the equation.
        :rtype: float'''
        if len(self.expression) == 0:
            self.infix_to_postfix()

        self.stack.clear()

        try:
            for token in self.expression:
                if isinstance(token, VariableToken):
                    if token.name in variables:
                        token.execute(variables[token.name])
                    elif token.name in Calculator.CONSTANTS:
                        token.execute(Calculator.CONSTANTS[token.name])
                    else:
                        raise VariableUndefinedError(token.name)
                else:
                    token.execute()
        except:
            pass
        else:
            if len(self.stack) > 0:
                result = self.stack[-1]
                if isinstance(result, (float, int)):
                    return result
        raise CalculationError(self.stream.text)



'''Module containing all tokens and classes needed to tokenize expressions.

:data TOKEN_REGEX: Regular Expression used for extracting specific token characters.
:type TOKEN_REGEX: Pattern'''

import re
from abc import ABC, abstractmethod
from typing import Callable, Dict, List, Pattern
from math import acos, acosh, asin, asinh, atan, atanh, cosh, degrees, e, fabs, fmod, pi, radians, sin, cos, sinh, tan, log, tanh

class Token(ABC):

    '''Token Abstract Base Class.'''

    def __init__(self, stack: List):
        self._stack = stack

    @abstractmethod
    def execute(self):
        '''Execute the function.'''
        pass

    @property
    def precedence(self):
        '''Precedence of a token.'''
        try:
            return self.__class__.PRECEDENCE
        except NameError: # If value isn't set
            return -1

    def __str__(self):
        return self.__class__.__name__

class VariableToken(Token):

    '''Variable token.'''

    def __init__(self, stack:List, name:str):
        super().__init__(stack)
        self.__name = name

    @property
    def name(self) -> str:
        '''Return the read-only name property.'''
        return self.__name

    def execute(self, value: float):
        '''Push given value to top of stack.

        :param value: What value to push.
        :type value: float'''
        self._stack.append(value)

    def __str__(self):
        return f"VariableToken({self.name})"

class ValueToken(Token):

    '''Token representing a number.'''

    def __init__(self, stack: List, value: float):
        super().__init__(stack)
        self.__value = value

    def execute(self):
        '''Push value to top of stack.'''
        self._stack.append(self.__value)
    
    def __str__(self):
        return f"ValueToken({self.__value})"

class TokenBuilder:

    '''Build tokens.'''

    TOKENS: Dict[str, Callable[[], Token]] = {}

    def __init__(self, stack: List):
        self._stack = stack

    def build_value(self, value: float) -> ValueToken:
        '''Build a value token.'''
        return ValueToken(self._stack, value)

    def build_variable(self, name: str) -> VariableToken:
        '''Build a variable token.'''
        return VariableToken(self._stack, name)

def add_builder(function_name:str, key: str):
    '''This decorator allows us to dynamically add tokens to the TokenBuilder.

    :param function_name: Name of the function to add.
    :type function_name: str
    :param key: What represents the function in an equation.
    :type key: str'''

    def decorate(cls):
        def func(self, *args, **kwargs):
            return cls(self._stack, *args, **kwargs)

        setattr(TokenBuilder, function_name, func)
        TokenBuilder.TOKENS[key] = function_name

        return cls

    return decorate
class BinaryToken(Token):

    '''Binary Operation token.'''

    @abstractmethod
    def operation(self, a, b):
        '''What operation to execute.'''
        pass

    def execute(self):
        '''Pop two items from the stack, run an operation on them,\
            and push the result to the stack.'''
        b = self._stack.pop()
        a = self._stack.pop()
        result = self.operation(a, b)
        self._stack.append(result)

@add_builder("build_add", "+")
class AddToken(BinaryToken):

    '''Addition token.'''

    PRECEDENCE = 1

    def operation(self, a, b):
        '''Add a and b'''
        return a + b

@add_builder("build_subtract", "-")
class SubtractToken(BinaryToken):

    '''Subtraction token.'''

    PRECEDENCE = 1

    def operation(self, a, b):
        '''Subtract a and'''
        return a - b

@add_builder("build_multiply", "*")
class MultiplyToken(BinaryToken):

    '''Multiplication token.'''

    PRECEDENCE = 2

    def operation(self, a, b):
        '''Multiply a and b'''
        return a * b

@add_builder("build_divide", "/")
class DivideToken(BinaryToken):

    '''Division token.'''

    PRECEDENCE = 2

    def operation(self, a, b):
        '''Divide a and b.'''
        return a / b

@add_builder("build_modulo", "%")
class ModuloToken(BinaryToken):

    '''Modulo token.'''

    PRECEDENCE = 2

    def operation(self, a, b):
        return fmod(a, b)

@add_builder("build_power", "^")
class PowerToken(BinaryToken):

    '''Exponation token.'''

    PRECEDENCE = 3

    def operation(self, a, b):
        '''Exponate a to the power of b'''
        return a ** b

class FunctionToken(Token):

    '''Token representing a unary function.'''

    @abstractmethod
    def operation(self, a):
        pass

    def execute(self):
        '''Pop an item from the stack, run a function on it,\
            and push the result to the stack.'''
        a = self._stack.pop()
        result = self.operation(a)
        self._stack.append(result)

@add_builder("build_sin", "sin(")
class SinToken(FunctionToken):

    '''Sine token.'''

    PRECEDENCE = 4

    def operation(self, a):
        '''Run sin'''
        return sin(a)

@add_builder("build_cos", "cos(")
class CosToken(FunctionToken):

    '''Cosine token.'''

    PRECEDENCE = 4

    def operation(self, a):
        '''Run cosine'''
        return cos(a)

@add_builder("build_tan", "tan(")
class TanToken(FunctionToken):

    '''Tangent token.'''

    PRECEDENCE = 4

    def operation(self, a):
        '''Run tangent'''
        return tan(a)

@add_builder("build_log", "log(")
class LogToken(FunctionToken):

    '''Logarithm token.'''

    PRECEDENCE = 4

    def operation(self, a):
        '''Run natural logarithm.'''
        return log(a)

@add_builder("build_sinh", "sinh(")
class SinhToken(FunctionToken):

    '''Hyperbolic Sine Token.'''

    PRECEDENCE = 4

    def operation(self, a):
        return sinh(a)

@add_builder("build_cosh", "cosh(")
class CoshToken(FunctionToken):

    '''Hyperbolic Cosine Token.'''

    PRECEDENCE = 4

    def operation(self, a):
        return cosh(a)

@add_builder("build_tanh", "tanh(")
class TanhToken(FunctionToken):

    '''Hyperbolic Tangent Token.'''

    PRECEDENCE = 4

    def operation(self, a):
        return tanh(a)

@add_builder("build_sign", "sign(")
class SignToken(FunctionToken):

    '''Sign Token.'''

    PRECEDENCE = 4

    def operation(self, a):
        if a < 0:
            return -1
        elif a > 0:
            return 1
        return 0

@add_builder("build_abs", "abs(")
class AbsToken(FunctionToken):

    '''Absolute Value token.'''

    PRECEDENCE = 4

    def operation(self, a):
        return fabs(a)

@add_builder("build_round", "round(")
class RoundToken(FunctionToken):

    '''Round Token.'''

    PRECEDENCE = 4

    def operation(self, a):
        return round(a)

@add_builder("build_rad", "rad(")
class RadiansToken(FunctionToken):

    '''Radian Conversion Token.'''

    PRECEDENCE = 4

    def operation(self, a):
        return radians(a)

@add_builder("build_deg", "deg(")
class DegreesToken(FunctionToken):

    '''Degrees Conversion Token.'''

    PRECEDENCE = 4

    def operation(self, a):
        return degrees(a)

@add_builder("build_acos", "acos(")
class ArcCosToken(FunctionToken):

    '''Arc Cosine token.'''

    PRECEDENCE = 4

    def operation(self, a):
        return acos(a)

@add_builder("build_asin", "asin(")
class ArcSinToken(FunctionToken):

    '''Arc Sine token.'''

    PRECEDENCE = 4

    def operation(self, a):
        return asin(a)

@add_builder("build_atan", "atan(")
class ArcTanToken(FunctionToken):

    '''Arc Tangent token.'''

    PRECEDENCE = 4

    def operation(self, a):
        return atan(a)

@add_builder("build_acosh", "acosh(")
class ArcCosHToken(FunctionToken):

    '''inverse hyperbolic cosine token.'''

    PRECEDENCE = 4

    def operation(self, a):
        return acosh(a)

@add_builder("build_asinh", "asinh(")
class ArcSinHToken(FunctionToken):

    '''inverse hyperbolic sine token.'''

    PRECEDENCE = 4

    def operation(self, a):
        return asinh(a)

@add_builder("build_atanh", "atanh(")
class ArcTanHToken(FunctionToken):

    '''inverse hyperbolic tangent token.'''

    PRECEDENCE = 4

    def operation(self, a):
        return atanh(a)

TOKEN_REGEX_STRING: str = r"\d+.?\d*|"
TOKEN_REGEX_STRING += "|".join([re.escape(token) for token in TokenBuilder.TOKENS.keys()]) 
TOKEN_REGEX_STRING += r"|\(|\)|[a-z]+"
TOKEN_REGEX: Pattern[str] = re.compile(TOKEN_REGEX_STRING) 

class TokenStream:

    '''A read-only representation of a stream of strings representing individual tokens.

    :property text: The infix represenetation of the full expression.
    :type text: str'''

    def __init__(self, expression: str):
        self.__expression: List[str] = TOKEN_REGEX.findall(expression)
        self.__iteration = 0

    @property
    def text(self) -> str:
        '''The infix representation of the full expression.'''
        return "".join(self.__expression)

    def reset(self):
        '''Reset the stream to the beginning.'''
        self.__iteration = 0

    def __iter__(self): 
        return self

    def __next__(self):
        if self.__iteration < len(self.__expression):
            value = self.__expression[self.__iteration]
            self.__iteration += 1
            return value
        raise StopIteration

    def __str__(self):
        return self.text



#!/bin/env python3
'''Setup script.'''
from pathlib import Path
from setuptools import setup, find_packages

THIS_DIRECTORY = Path(__file__).parent

REQUIREMENTS = (THIS_DIRECTORY / "requirements.txt").read_text().split('\n')[:-1]
LONG_DESCRIPTION = (THIS_DIRECTORY / "README.md").read_text()

CONTENT = {
    "name": "veq",
    "version": "1.3.4",
    "author": "Willow Ciesialka",
    "author_email": "wciesialka@gmail.com",
    "url": "https://github.com/wciesialka/riemann-sum-py",
    "description": "Equation visualizer.",
    "long_description": LONG_DESCRIPTION,
    "long_description_content_type": "text/markdown",
    "license": "GPL-3.0",
    "packages": find_packages(where="src"),
    "entry_points": {
        'console_scripts': [
            'veq = veq.__main__:main'
        ]
    },
    "classifiers": [
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Programming Language :: Python :: 3",
        "Topic :: Artistic Software",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Mathematics"
    ],
    "keywords": "python image math",
    "package_dir": {"": "src"},
    "install_requires": REQUIREMENTS,
    "zip_safe": False,
    "python_requires": ">=3.8.10"
}

setup(**CONTENT)



'''Equation visualization module.'''

import logging
from typing import Tuple, List, Union
from time import time
import pygame
from veq.calculator import CalculationError, Calculator

def remap(value: float, input_range: Tuple[float, float], output_range: Tuple[float, float]) -> float:
    '''Map a value from an input range to an output range.

    :param value: Value to map.
    :type value: float
    :param input_range: Input range to map from.
    :type input_range: tuple(float, float)
    :param output_range: Output range to map to.
    :type output_range: tuple(float, float)
    :returns: Mapped value.
    :rtype: float
    '''
    # f(x) = o_s + ((o_e - o_s) / (i_e - i_s)) * (x - i_s)
    input_start, input_end = input_range
    output_start, output_end = output_range
    divisor = input_end - input_start
    dividend = output_end - output_start
    multiplier = value - input_start

    return output_start + (dividend / divisor) * multiplier

class Equation:

    '''An equation with an expression, domain, and range.

    :attribute calculator: The calculator holding the postfix expression to calculate.
    :type calculator: veq.calculator.Calculator
    :attribute domain: Domain of the equation.
    :type domain: tuple(float, float)
    :attribute range: Range of the equation.
    :type range: tuple(float, float)'''

    def __init__(self, calculator: Calculator, \
        domain: Tuple[float, float], range: Tuple[float, float]):

        self.domain = domain
        self.range = range
        self.calculator = calculator
        self.calculator.infix_to_postfix()

    def zoom(self, increase: float):
        '''Adjust the domain and range by a given number to give the illusion of zooming in/out.

        :param increase: Number to adjust by. Half will be subtracted\
            from the left bound of the domain and range, while \
                another half will be added to the right bound.
        :type increase: float'''
        if (self.range[0] - increase/2 < self.range[1] + increase/2) and\
             (self.domain[0] - increase/2 < self.domain[1] + increase/2):
            self.range[0] -= increase/2
            self.range[1] += increase/2
            self.domain[0] -= increase/2
            self.domain[1] += increase/2

    def shift(self, shift_by: Tuple[float, float]):
        '''Shift the domain and range of the equation.

        :param shift_by: What values to adjust by.
        :type shift_by: tuple(float, float)'''
        self.domain[0] += shift_by[0]
        self.domain[1] += shift_by[0]
        self.range[0] += shift_by[1]
        self.range[1] += shift_by[1]

class Visualizer:

    '''Graphical visualization of an equation.

    :attribute equation: What equation to plot.
    :type equation: veq.visualizer.Equation
    :attribute screen: pygame surface to plot on.
    :type screen: pygame.Surface
    :attribute color: What color of line to plot.
    :type color: tuple(int, int, int)
    :value color: tuple(0, 0, 0)
    :property width: Width of the screen in pixels.
    :type width: int
    :property height: Height of the screen in pixels.
    :type height: int
    :property left: Left bound of the equation's domain.
    :type left: float
    :property right: Right bound of the equation's domain.
    :type right: float
    :property top: Left bound of the equation's range.
    :type top: float
    :property bottom: Right bound of the equation's range.
    :type bottom: float'''

    def __init__(self, equation: Equation, screen: pygame.Surface, *,\
         color: Tuple[int, int, int] = (0, 0, 0)):

        pygame.font.init()
        self.__font = pygame.font.SysFont("Courier New", 12)
        self.equation = equation
        self.screen = screen
        self.color = color
        self.__saved = None

        self.__t = 0
        self.__last_render = 0

    @property
    def left(self):
        '''Left bound of the equation.'''
        return self.equation.domain[0]

    @property
    def right(self):
        '''Right bound of the equation.'''
        return self.equation.domain[1]

    def execute(self, x: float) -> Union[float, None]:
        '''Calculate the Calculator's expression with value x.

        :param x: Value to substitute x for in the Visualizer's equation.
        :type x: float
        :returns: The resulting expression, or None if there is no valid\
             calculation for that x value.
        :rtype: float or None'''
        try:
            y = self.equation.calculator.calculate(x = x, t = self.__t)
        except CalculationError:
            return None
        else:
            return y

    @property
    def width(self):
        '''Width of the screen.'''
        return self.screen.get_width()

    @property
    def height(self):
        '''Height of the screen.'''
        return self.screen.get_height()

    @property
    def top(self):
        '''Top bound of the equation.'''
        return self.equation.range[1]

    @property
    def bottom(self):
        '''Bottom bound of the equation.'''
        return self.equation.range[0]

    def reset_t(self):
        '''Reset the internal t variable.'''
        self.__last_render = 0
        self.__t = 0

    def draw_equation(self):
        '''Plot the equation to the screen.'''
        render_start = time()

        dx: float = (self.right-self.left)/self.width

        list_of_points = []
        points: List[Tuple[int, float]] = []
        for i in range(0, self.width):
            # Get X and Y
            x = self.left + (dx*i)
            y = self.execute(x)
            if y is None:
                continue
            else:
                screen_y = remap(y, (self.bottom, self.top), (self.height, 0))

                # If considerably enough on screen...
                if abs(screen_y) > self.height * 4:
                    list_of_points.append(points)
                    points = []

                point: Tuple[int, float] = (i, screen_y)
                points.append(point)

        list_of_points.append(points)

        for points in list_of_points:
            if len(points) >= 2:
                try:
                    pygame.draw.lines(self.screen, self.color, False, points, 1)
                except Exception as ex:
                    logging.debug(points)
                    raise ex

        current_time = time()
        if self.__last_render == 0:
            self.__t = current_time - render_start
        else:
            self.__t += current_time - self.__last_render
        self.__last_render = current_time

    def draw_text(self):
        '''Draw text onto the screen.'''

        domain_text = f'Domain: [{self.left:1.2f}, {self.right:1.2f}]'
        range_text = f'Range: [{self.bottom:1.2f}, {self.top:1.2f}]'
        equation_text = f"Equation: {self.equation.calculator.stream}"
        t_text = f"t: {self.__t:1.2f}"

        domain_surface = self.__font.render(domain_text, True, (0, 0, 0), (255, 255, 255))
        range_surface = self.__font.render(range_text, True, (0, 0, 0), (255, 255, 255))
        equation_surface = self.__font.render(equation_text, True, (0, 0, 0), (255, 255, 255))

        t_surface = self.__font.render(t_text, True, (0, 0, 0), (255, 255, 255))
        t_width = t_surface.get_size()[0]

        self.screen.blit(domain_surface, (0, 12))
        self.screen.blit(range_surface, (0, 24))
        self.screen.blit(equation_surface, (0, 0))
        self.screen.blit(t_surface, (self.width - t_width, 0))

    def __domain_range(self, step):
        x = (self.left // step) * step
        while x < self.right:
            x += step
            yield x
        return x

    def __range_range(self, step):
        y = (self.bottom // step) * step
        while y < self.top:
            y += step
            yield y
        return y

    def draw_grid(self, step: float):
        '''Draw a grid onto the screen, with the lines seperated\
             by a certain step.

        :param step: What amount to seperate gridlines by.
        :type step: float'''

        if round(step) == step:
            step_precision = 0
        else:
            step_str = str(step)
            step_precision = len(step_str[step_str.index('.')+1:])

        text_format = lambda x: str(round(x, step_precision))

        grid_color = (200, 200, 200)
        text_color = (100, 100, 100)

        for x in self.__domain_range(step):
            screen_x = remap(x, (self.left, self.right), (0, self.width))
            pygame.draw.line(self.screen, grid_color, (screen_x, 0), (screen_x, self.height))
            text = text_format(x)
            text_surface = self.__font.render(text, True, text_color, (255, 255, 255))
            self.screen.blit(text_surface, (screen_x - (len(text)/2)*6, self.height-12))

        for y in self.__range_range(step):
            screen_y = remap(y, (self.bottom, self.top), (self.height, 0))
            pygame.draw.line(self.screen, grid_color, (0, screen_y), (self.width, screen_y))
            text = text_format(y)
            text_surface = self.__font.render(text, True, text_color, (255, 255, 255))
            self.screen.blit(text_surface, (12, screen_y - 6))

    def draw_axis(self):
        '''Draw the axis lines onto the screen.'''

        axis_color = (205, 25, 25)
        screen_x = remap(0, (self.left, self.right), (0, self.width))
        screen_y = remap(0, (self.bottom, self.top), (self.height, 0))
        if 0 <= screen_x <= self.width:
            pygame.draw.line(self.screen, axis_color, (screen_x, 0), (screen_x, self.height))
        if 0 <= screen_y <= self.height:
            pygame.draw.line(self.screen, axis_color, (0, screen_y), (self.width, screen_y))

    def draw_location(self, mouse_pos: Tuple[int, int]):
        '''Draw the location text information onto the screen.

        :param mouse_pos: Position of the mouse on the screen.
        :type mouse_pos: tuple(int, int)'''

        color = (0, 0, 175)

        mouse_x, mouse_y = mouse_pos
        x = remap(mouse_x, (0, self.width), (self.left, self.right))
        y = self.execute(x)
        if not y is None:
            screen_y = remap(y, (self.bottom, self.top), (self.height, 0))
            location_text = f"Location: ({x:.2f}, {y:.2f})"

            pygame.draw.circle(self.screen, color, (mouse_x, screen_y), 5, width=2)
            text_surface = self.__font.render(location_text, True, (0, 0, 0), (255, 255, 255))
            self.screen.blit(text_surface, (0, 48))

    def save(self, mouse_pos: Tuple[int, int]):
        '''Save the current location into the visualizer.

        :param mouse_pos: Position of the mouse on the screen.
        :type mouse_pos: tuple(int, int)'''
        mouse_x, mouse_y = mouse_pos
        x = remap(mouse_x, (0, self.width), (self.left, self.right))
        self.__saved = x

    def on_screen(self, pixel: Tuple[int, int]) -> bool:
        '''Returns whether the given pixel is within the bounds of the screen.

        :param pixel: What pixel to check.
        :type pixel: tuple(int, int)
        :returns: true if the pixel is within the bounds of the screen, and false if it isn't.
        :rtype: bool'''
        return 0 <= pixel[0] <= self.width and 0 <= pixel[1] <= self.height

    def draw_saved(self):
        '''Draw the saved location onto the screen.'''
        color = (25, 175, 25)

        if not self.__saved is None:
            x = self.__saved
            y = self.execute(x)
            if not y is None:
                screen_x = remap(x, (self.left, self.right), (0, self.width))
                screen_y = remap(y, (self.bottom, self.top), (self.height, 0))
                saved_text = f"Saved: ({x:.2f}, {y:.2f})"
                text_surface = self.__font.render(saved_text, True, (0, 0, 0), (255, 255, 255))
                self.screen.blit(text_surface, (0, 36))
                if self.on_screen((screen_x, screen_y)):
                    pygame.draw.circle(self.screen, color, (screen_x, screen_y), 4, width=1)
